@@ -10,11 +10,12 @@ use warnings;
 use feature 'current_sub';
 use experimental 'signatures', 'lexical_subs', 'postderef';
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 use base 'Exporter::Tiny';
 use Hash::Util 'lock_hash', 'unlock_hash', 'unlock_value';
-use Scalar::Util 'refaddr';
+use POSIX 'floor';
+use Scalar::Util 'refaddr', 'looks_like_number';
 use Ref::Util 'is_plain_arrayref', 'is_plain_hashref', 'is_coderef', 'is_plain_refref';
 use Feature::Compat::Defer;
 
@@ -24,15 +25,15 @@ use constant {
     CYCLE_TRUNCATE => 'truncate',
 };
 use constant CYCLE_QR => qr /\A die|continue|truncate \z/x;
-use constant { VISIT_CONTAINER => 0b01, VISIT_LEAF => 0b10, VISIT_ALL => 0b11 };
+use constant { VISIT_CONTAINER => 0b01, VISIT_LEAF      => 0b10, VISIT_ALL                => 0b11 };
+use constant { RESULT_RETURN   => 0,    RESULT_CONTINUE => 1,    RESULT_REVISIT_CONTAINER => 2 };
 
 our %EXPORT_TAGS = (
     funcs     => [qw( visit )],
-    constants => [ qw(
-          VISIT_CONTAINER VISIT_LEAF VISIT_ALL
-          CYCLE_DIE CYCLE_CONTINUE CYCLE_TRUNCATE
-        ),
-    ],
+    results   => [qw( RESULT_RETURN RESULT_CONTINUE RESULT_REVISIT_CONTAINER )],
+    cycles    => [qw( CYCLE_DIE CYCLE_CONTINUE CYCLE_TRUNCATE )],
+    visits    => [qw( VISIT_CONTAINER VISIT_LEAF VISIT_ALL )],
+    constants => [qw( :results :cycles :visits )],
 );
 
 our @EXPORT_OK = map { $EXPORT_TAGS{$_}->@* } keys %EXPORT_TAGS;
@@ -45,8 +46,9 @@ my sub croak {
 
 my sub _visit ( $node, $code, $context, $cycle, $visit, $meta ) {    ## no critic (Subroutines::ProhibitManyArgs)
 
-    my $path      = $meta->{path};
-    my $ancestors = $meta->{ancestors};
+    my $path          = $meta->{path};
+    my $ancestors     = $meta->{ancestors};
+    my $revisit_limit = $meta->{revisit_limit};
 
     my $refaddr = refaddr( $node );
     if ( exists $meta->{seen}{$refaddr} ) {
@@ -63,12 +65,10 @@ my sub _visit ( $node, $code, $context, $cycle, $visit, $meta ) {    ## no criti
         $lcycle eq CYCLE_CONTINUE
           or croak( __PACKAGE__ . "::visit: unkown cycle parameter value: $lcycle" );
     }
-    else {
-        $meta->{seen}{$refaddr} = $node;
-    }
 
     # after this call to _visit, will have visited all descendents of
     # $node, so don't need this any longer.
+    $meta->{seen}{$refaddr} = ();
     defer { delete $meta->{seen}{$refaddr} }
 
     my %meta = $meta->%*;
@@ -79,33 +79,44 @@ my sub _visit ( $node, $code, $context, $cycle, $visit, $meta ) {    ## no criti
 
     my $is_hashref = is_plain_hashref( $node );
 
-    my @idx = $is_hashref ? sort keys $node->%* : keys $node->@*;
-
     push $ancestors->@*, $node;
     defer { pop $ancestors->@* };
 
     my $visit_leaf = !!( $visit & VISIT_LEAF );
     my $visit_node = !!( $visit & VISIT_CONTAINER );
 
-    for my $idx ( @idx ) {
+  SCAN: {
+        last unless --$revisit_limit;
 
-        push $path->@*, $idx;
-        defer { pop $path->@* }
+        my @idx = $is_hashref ? sort keys $node->%* : keys $node->@*;
 
-        my $vref = \( $is_hashref ? $node->{$idx} : $node->[$idx] );
+        for my $idx ( @idx ) {
 
-        my $is_node = is_plain_refref( $vref );
+            push $path->@*, $idx;
+            defer { pop $path->@* }
 
-        my $visit_element = $is_node ? $visit_node : $visit_leaf;
+            my $vref = \( $is_hashref ? $node->{$idx} : $node->[$idx] );
 
-        return !!0 if $visit_element && !$code->( $idx, $vref, $context, \%meta );
+            my $is_node = is_plain_refref( $vref );
+
+            my $visit_element = $is_node ? $visit_node : $visit_leaf;
+
+            if ( $visit_element
+                and ( my $result = $code->( $idx, $vref, $context, \%meta ) ) != RESULT_CONTINUE )
+            {
+                redo SCAN  if $result == RESULT_REVISIT_CONTAINER;
+                return !!0 if $result == RESULT_RETURN;
+            }
+
+            next unless is_plain_refref( $vref );
 
-        if ( is_plain_refref( $vref ) ) {
             my $ref = $vref->$*;
             __SUB__->( $ref, $code, $context, $cycle, $visit, \%meta ) || return !!0
               if is_plain_arrayref( $ref ) || is_plain_hashref( $ref );
         }
     }
+    croak( "exceeded limit ($meta->{revisit_limit}) on revisiting containers" )
+      unless $revisit_limit;
 
     return !!1;
 }
@@ -279,6 +290,48 @@ my sub _visit ( $node, $code, $context, $cycle, $visit, $meta ) {    ## no criti
 
 
 
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
 
 
 
@@ -310,12 +363,18 @@ sub visit ( $struct, $callback, %opts ) {
       or croak( q{parameter 'context' must be a plain hash} );
 
     my %metadata = (
-        path      => [],
-        seen      => {},
-        ancestors => [],
-        container => undef,
+        path          => [],
+        seen          => {},
+        ancestors     => [],
+        container     => undef,
+        revisit_limit => delete $opts{revisit_limit} // 10,
     );
 
+    croak( "illegal value for 'revisit_limit' : $metadata{revisit_limit}" )
+      unless looks_like_number( $metadata{revisit_limit} )
+      && floor( $metadata{revisit_limit} ) == $metadata{revisit_limit};
+
+
     my $cycle = delete $opts{cycle} // 'die';
     my $visit = delete $opts{visit} // VISIT_ALL;
 
@@ -361,11 +420,11 @@ CXC::Data::Visitor - Invoke a callback on every element at every level of a data
 
 =head1 VERSION
 
-version 0.02
+version 0.03
 
 =head1 SYNOPSIS
 
-  use CXC::Data::Visitor 'visit';
+  use CXC::Data::Visitor 'visit', 'RESULT_CONTINUE';
 
   my $hoh = { fruit => { berry => 'purple' }, };
 
@@ -374,6 +433,7 @@ version 0.02
       sub {
           my ( $key, $vref ) = @_;
           $$vref = 'blue' if $key eq 'berry';
+          return RESULT_CONTINUE;
       } );
 
   say $hoh->{fruit}{berry}    # 'blue'
@@ -392,6 +452,11 @@ The traversal may be aborted.
 
 =item *
 
+A container (hash or array) may be revisited if the callback changes
+it.
+
+=item *
+
 User selectable behavior upon detection of a traversal cycle.
 
 =item *
@@ -408,27 +473,51 @@ path) is available.
 
    ( $completed, $context, $metadata ) = visit( $struct, $callback, %opts );
 
-Perform a depth-first traversal of B<$struct>, invoke B<$callback> on
-every element (hash entry or array element) in
-B<$struct>. B<$callback> is invoked on a container first before being
-called on its elements.  Blessed hashes or arrays are not traversed.
-Cycles are detected, but nodes which are accessible from multiple
-parents are visited more than once.
+Perform a depth-first traversal of B<$struct>, invoking B<$callback>
+on containers (hashes and arrays) and terminal elements in B<$struct>.
 
-For example, if
+=over
 
-  $struct = { a => { b => [ 0, 1 ], c => 2 } };
+=item *
 
-The calling order is
+Blessed hashes or arrays are treated as terminal elements and are not
+further traversed.
 
-  $struct->{a}
-  $struct->{a}{b}
-  $struct->{a}{b}[0]
-  $struct->{a}{b}[1]
-  $struct->{a}{b}[2]
-  $struct->{a}{c}
+=item *
 
-The values returned are:
+Hashes are traversed in sorted key order.
+
+=item *
+
+Cycles are detected upon traversing a container a second time in a
+depth first search.
+
+=item *
+
+Containers that can be reached multiple times without cycling, e.g.
+
+  %hash = ( a => { b => 1 }, );
+  $hash{c} = $hash{a};
+
+are visited once per parent, e.g.
+
+  {a}, {a}{b}, {a}{b}[0]
+  {c}, {c}{b}, {c}{b}[0]
+
+=item *
+
+B<$callback> is first invoked on a container and then its
+elements.  Given
+
+  { a => { b => [ 0 ], c => 2 } }
+
+the callback order is
+
+  {a}, {a}{b}, {a}{b}[0], {a}{c}
+
+=back
+
+L</visit> returns the following:
 
 =over
 
@@ -447,109 +536,36 @@ collected metadata. See L</$metadata> below.
 
 =back
 
-B<%opts> may contain the following entries:
-
-=over
-
-=item B<context> => I<hashref>
-
-A reference to a hash which is passed to L</$callback>. It defaults to
-a freshly created hash.
-
-=item B<cycle> => CYCLE_TRUNCATE | CYCLE_DIE | CYCLE_CONTINUE | <$coderef>
-
-How to handle cycles in the data structure (the constants are
-available for import; see L</EXPORTS>).  Cycles are detected upon
-traversing a node a second time in a depth first search.
-
-Note that a node can be reached multiple times without cycling, for
-example, this is not a cycle:g
-
-  %hash = ( a => { b => [ 0, 1 ] }, );
-  $hash{c} = $hash{a};
-
-Acceptable values are:
-
-=over
-
-=item CYCLE_DIE
-
-Throw an exception (the default).
-
-=item CYCLE_CONTINUE
-
-Pretend we haven't seen it before. Will cause stack exhaustion if
-B<$callback> does handle this.
-
-=item CYCLE_TRUNCATE
-
-Truncate before entering the cycle a second time.
-
-=item I<$coderef>
-
-Examine the situation and request a particular resolution.
-B<$coderef> is called as
-
-  $cycle = $coderef->( $container, $context, $metadata );
-
-where B<$container> is the hash or array which has already been
-traversed. See below for L</$context> and L</$metadata>.
-
-B<$cycle> should be one of B<CYCLE_DIE>, B<CYCLE_CONTINUE>, or B<CYCLE_TRUNCATE>,
-and indicates what should be done.
-
-=back
-
-=item I<visit> VISIT_CONTAINER, VISIT_LEAF, VISIT_ALL
-
-The parts of the structure will trigger a callback
-(the constants are available for import; see L</EXPORTS>).
-
-Acceptable values are:
-
-=over
-
-=item VISIT_CONTAINER
-
-Only containers get a callback.  For example, the elements in the following structure
-
-  $struct = { a => { b => 1, c => [ 2, 3 ] } }
-
-which will be passed to L</$callback> are:
-
-  a => {...}  # $struct->{a}
-  c => [...]  # $struct->{c}
-
-=item VISIT_LEAF
-
-Only containers get a callback.  For example, the elements in the following structure
-
-  For example, the following structure
-
-  $struct = { a => { b => 1, c => [ 2, 3 ] } }
-
-which will be passed to L</$callback> are:
-
-  b => 1  # $struct->{a}{b}
-  0 => 2  # $struct->{a}{c}[0]
-  1 => 3  # $struct->{a}{c}[1]
-
-=item VISIT_ALL
-
-All elements get a callback. I<Default>
-
-=back
-
-=back
-
 B<$callback> will be called as:
 
-  $continue = $callback->( $kydx, $vref, $context, $metadata );
+  $handle_return = $callback->( $kydx, $vref, $context, $metadata );
 
-and should return B<true> if the process of visiting elements should
-I<continue>, B<false> if L</visit> should return immediately.
+and should return one of the following (see L</EXPORTS> to import the constants):
 
-Its arguments are
+=over
+
+=item RESULT_CONTINUE
+
+The process of visiting elements should continue.
+
+=item RESULT_RETURN
+
+L</visit> should return immediately to the caller.
+
+=item RESULT_REVISIT_CONTAINER
+
+Further processing of the elements in the current container should stop
+and the container should be revisited.  This allows L</$callback> to
+modify the container and have it reprocessed.
+
+To avoid inadvertent infinite loops, a finite number of revisits
+is allowed during a traversal of a container (see L</revisit_limit>).
+Containers with multiple parents are traversed once per parent;
+The limit is reset for each traversal.
+
+=back
+
+L</$callback> is passed
 
 =over
 
@@ -590,6 +606,98 @@ An array contains the ancestor containers of the current element.
 
 =back
 
+B<%opts> may contain the following entries:
+
+=over
+
+=item B<context> => I<hashref>
+
+A reference to a hash passed to L</$callback>, for the caller's
+use. It defaults to a freshly created hash.
+
+=item B<cycle> => CYCLE_TRUNCATE | CYCLE_DIE | CYCLE_CONTINUE | <$coderef>
+
+How to handle cycles in the data structure.
+
+See L</EXPORTS> to import the constant values.
+
+=over
+
+=item CYCLE_DIE
+
+Throw an exception (the default).
+
+=item CYCLE_CONTINUE
+
+Pretend we haven't seen it before. Will cause stack exhaustion if
+B<$callback> does handle this.
+
+=item CYCLE_TRUNCATE
+
+Truncate before entering the cycle a second time.
+
+=item I<$coderef>
+
+Examine the situation and request a particular resolution.
+B<$coderef> is called as
+
+  $coderef->( $container, $context, $metadata );
+
+where B<$container> is the hash or array which has already been
+traversed. See below for L</$context> and L</$metadata>.
+
+B<$coderef> should return one of B<CYCLE_DIE>, B<CYCLE_CONTINUE>, or B<CYCLE_TRUNCATE>,
+indicating what should be done.
+
+=back
+
+=item I<visit> => VISIT_CONTAINER | VISIT_LEAF | VISIT_ALL
+
+The parts of the structure that will trigger a callback.
+See L</EXPORTS> to import the constants.
+
+=over
+
+=item VISIT_CONTAINER
+
+Invoke L</$callback> on containers.  For example, the elements in the
+following structure
+
+  $struct = { a => { b => 1, c => [ 2, 3 ] } }
+
+passed to L</$callback> are:
+
+  a => {...}  # $struct->{a}
+  c => [...]  # $struct->{c}
+
+=item VISIT_LEAF
+
+Invoke L</$callback> on terminal (leaf) elements.  For example, the
+elements in the following structure
+
+  $struct = { a => { b => 1, c => [ 2, 3 ] } }
+
+passed to L</$callback> are:
+
+  b => 1  # $struct->{a}{b}
+  0 => 2  # $struct->{a}{c}[0]
+  1 => 3  # $struct->{a}{c}[1]
+
+=item VISIT_ALL
+
+Invoke L</$callback> on all elements.  This is the default.
+
+=back
+
+=item revisit_limit
+
+If L</$callback> returns B<RESULT_REVISIT_CONTAINER>, then the parent container
+of the element is re-scanned for its elements and revisited.  To avoid an inadvertent
+infinite loop, an exception is thrown if the parent container is revisited more
+than this number of times.  It defaults to 10;
+
+=back
+
 =head1 EXPORTS
 
 This module uses L<Exporter::Tiny>, which provides enhanced import utilities.
@@ -599,21 +707,31 @@ The following symbols may be exported:
   visit
   VISIT_CONTAINER VISIT_LEAF VISIT_ALL
   CYCLE_DIE CYCLE_CONTINUE CYCLE_TRUNCATE
+  RESULT_RETURN RESULT_CONTINUE RESULT_REVISIT_CONTAINER
 
-The following export tags are available:
+The available tags and their respective imported symbols are:
 
 =over
 
-=item B<-all>
+=item B<all>
 
-Import all symbols
+Import all symbols.
 
-=item B<-constants>
+=item B<results>
 
-Import the following symbols
+ RESULT_RETURN RESULT_CONTINUE RESULT_REVISIT_CONTAINER
 
-  VISIT_CONTAINER VISIT_LEAF VISIT_ALL
-  CYCLE_DIE CYCLE_CONTINUE CYCLE_TRUNCATE
+=item B<cycles>
+
+ CYCLE_DIE CYCLE_CONTINUE CYCLE_TRUNCATE
+
+=item B<visits>
+
+ VISIT_CONTAINER VISIT_LEAF VISIT_ALL
+
+=item B<constants>
+
+Import tags C<results>, C<cycles>, C<visits>.
 
 =back
 
